@@ -42,6 +42,7 @@ const char* GetSectionName(BinarySection sec) {
   }
 }
 
+// clang-format off
 const char* NameSubsectionName[] = {
   "module",
   "function",
@@ -53,9 +54,14 @@ const char* NameSubsectionName[] = {
   "global",
   "elemseg",
   "dataseg",
+  "tag",
 };
+// clang-format on
 
 const char* GetNameSectionSubsectionName(NameSectionSubsection subsec) {
+  static_assert(WABT_ENUM_COUNT(NameSectionSubsection) ==
+                    WABT_ARRAY_SIZE(NameSubsectionName),
+                "Malformed ExprTypeName array");
   return NameSubsectionName[size_t(subsec)];
 }
 
