@@ -78,6 +78,7 @@ enum class ObjectKind {
   Null,
   Foreign,
   Trap,
+  Exception,
   DefinedFunc,
   HostFunc,
   Table,
@@ -90,7 +91,7 @@ enum class ObjectKind {
 };
 
 const char* GetName(Mutability);
-const char* GetName(ValueType);
+const std::string GetName(ValueType);
 const char* GetName(ExternKind);
 const char* GetName(ObjectKind);
 
@@ -305,6 +306,32 @@ struct LocalDesc {
   u32 end;
 };
 
+// Metadata for representing exception handlers associated with a function's
+// code. This is needed to look up exceptions from call frames from interpreter
+// instructions.
+struct CatchDesc {
+  Index tag_index;
+  u32 offset;
+};
+
+// Handlers for a catch-less `try` or `try-catch` block are included in the
+// Catch kind. `try-delegate` instructions create a Delegate handler.
+enum class HandlerKind { Catch, Delegate };
+
+struct HandlerDesc {
+  HandlerKind kind;
+  u32 try_start_offset;
+  u32 try_end_offset;
+  std::vector<CatchDesc> catches;
+  union {
+    u32 catch_all_offset;
+    u32 delegate_handler_index;
+  };
+  // Local stack heights at the handler site that need to be restored.
+  u32 values;
+  u32 exceptions;
+};
+
 struct FuncDesc {
   // Includes params.
   ValueType GetLocalType(Index) const;
@@ -312,6 +339,7 @@ struct FuncDesc {
   FuncType type;
   std::vector<LocalDesc> locals;
   u32 code_offset;
+  std::vector<HandlerDesc> handlers;
 };
 
 struct TableDesc {
@@ -378,13 +406,19 @@ struct ModuleDesc {
 //// Runtime ////
 
 struct Frame {
-  explicit Frame(Ref func, u32 values, u32 offset, Instance*, Module*);
+  explicit Frame(Ref func,
+                 u32 values,
+                 u32 exceptions,
+                 u32 offset,
+                 Instance*,
+                 Module*);
 
   void Mark(Store&);
 
   Ref func;
   u32 values;  // Height of the value stack at this activation.
-  u32 offset;  // Istream offset; either the return PC, or the current PC.
+  u32 exceptions;  // Height of the exception stack at this activation.
+  u32 offset;      // Istream offset; either the return PC, or the current PC.
 
   // Cached for convenience. Both are null if func is a HostFunc.
   Instance* inst;
@@ -461,6 +495,7 @@ class Store {
   ObjectList::Index object_count() const;
 
   const Features& features() const;
+  void setFeatures(const Features& features) { features_ = features; }
 
  private:
   template <typename T>
@@ -646,6 +681,27 @@ class Trap : public Object {
   std::vector<Frame> trace_;
 };
 
+class Exception : public Object {
+ public:
+  static bool classof(const Object* obj);
+  static const ObjectKind skind = ObjectKind::Exception;
+  static const char* GetTypeName() { return "Exception"; }
+  using Ptr = RefPtr<Exception>;
+
+  static Exception::Ptr New(Store&, Ref tag, Values& args);
+
+  Ref tag() const;
+  Values& args();
+
+ private:
+  friend Store;
+  explicit Exception(Store&, Ref, Values&);
+  void Mark(Store&) override;
+
+  Ref tag_;
+  Values args_;
+};
+
 class Extern : public Object {
  public:
   static bool classof(const Object* obj);
@@ -1024,6 +1080,7 @@ enum class RunResult {
   Ok,
   Return,
   Trap,
+  Exception,
 };
 
 // TODO: Kinda weird to have a thread as an object, but it makes reference
@@ -1182,12 +1239,18 @@ class Thread : public Object {
   template <typename T, typename V = T>
   RunResult DoAtomicRmwCmpxchg(Instr, Trap::Ptr* out_trap);
 
+  RunResult DoThrow(Exception::Ptr exn_ref);
+
   RunResult StepInternal(Trap::Ptr* out_trap);
 
   std::vector<Frame> frames_;
   std::vector<Value> values_;
   std::vector<u32> refs_;  // Index into values_.
 
+  // Exception handling requires tracking a separate stack of caught
+  // exceptions for catch blocks.
+  RefVec exceptions_;
+
   // Cached for convenience.
   Store& store_;
   Instance* inst_ = nullptr;
