@@ -36,10 +36,10 @@ const char* ExprTypeName[] = {
   "Block",
   "Br",
   "BrIf",
-  "BrOnExn",
   "BrTable",
   "Call",
   "CallIndirect",
+  "CallRef",
   "Compare",
   "Const",
   "Convert",
@@ -68,8 +68,11 @@ const char* ExprTypeName[] = {
   "ReturnCallIndirect",
   "Select",
   "SimdLaneOp",
+  "SimdLoadLane",
+  "SimdStoreLane",
   "SimdShuffleOp",
   "LoadSplat",
+  "LoadZero",
   "Store",
   "TableCopy",
   "ElemDrop",
@@ -132,8 +135,8 @@ Index Module::GetFuncTypeIndex(const Var& var) const {
   return type_bindings.FindIndex(var);
 }
 
-Index Module::GetEventIndex(const Var& var) const {
-  return event_bindings.FindIndex(var);
+Index Module::GetTagIndex(const Var& var) const {
+  return tag_bindings.FindIndex(var);
 }
 
 Index Module::GetDataSegmentIndex(const Var& var) const {
@@ -158,8 +161,8 @@ bool Module::IsImport(ExternalKind kind, const Var& var) const {
     case ExternalKind::Table:
       return GetTableIndex(var) < num_table_imports;
 
-    case ExternalKind::Event:
-      return GetEventIndex(var) < num_event_imports;
+    case ExternalKind::Tag:
+      return GetTagIndex(var) < num_tag_imports;
 
     default:
       return false;
@@ -274,12 +277,12 @@ Memory* Module::GetMemory(const Var& var) {
   return memories[index];
 }
 
-Event* Module::GetEvent(const Var& var) const {
-  Index index = GetEventIndex(var);
-  if (index >= events.size()) {
+Tag* Module::GetTag(const Var& var) const {
+  Index index = GetTagIndex(var);
+  if (index >= tags.size()) {
     return nullptr;
   }
-  return events[index];
+  return tags[index];
 }
 
 const DataSegment* Module::GetDataSegment(const Var& var) const {
@@ -357,12 +360,12 @@ void Module::AppendField(std::unique_ptr<ElemSegmentModuleField> field) {
   fields.push_back(std::move(field));
 }
 
-void Module::AppendField(std::unique_ptr<EventModuleField> field) {
-  Event& event = field->event;
-  if (!event.name.empty()) {
-    event_bindings.emplace(event.name, Binding(field->loc, events.size()));
+void Module::AppendField(std::unique_ptr<TagModuleField> field) {
+  Tag& tag = field->tag;
+  if (!tag.name.empty()) {
+    tag_bindings.emplace(tag.name, Binding(field->loc, tags.size()));
   }
-  events.push_back(&event);
+  tags.push_back(&tag);
   fields.push_back(std::move(field));
 }
 
@@ -448,13 +451,13 @@ void Module::AppendField(std::unique_ptr<ImportModuleField> field) {
       break;
     }
 
-    case ExternalKind::Event: {
-      Event& event = cast<EventImport>(import)->event;
-      name = &event.name;
-      bindings = &event_bindings;
-      index = events.size();
-      events.push_back(&event);
-      ++num_event_imports;
+    case ExternalKind::Tag: {
+      Tag& tag = cast<TagImport>(import)->tag;
+      name = &tag.name;
+      bindings = &tag_bindings;
+      index = tags.size();
+      tags.push_back(&tag);
+      ++num_tag_imports;
       break;
     }
   }
@@ -532,8 +535,8 @@ void Module::AppendField(std::unique_ptr<ModuleField> field) {
       AppendField(cast<StartModuleField>(std::move(field)));
       break;
 
-    case ModuleFieldType::Event:
-      AppendField(cast<EventModuleField>(std::move(field)));
+    case ModuleFieldType::Tag:
+      AppendField(cast<TagModuleField>(std::move(field)));
       break;
   }
 }
@@ -661,11 +664,12 @@ uint8_t ElemSegment::GetFlags(const Module* module) const {
       break;
   }
 
-  all_ref_func = all_ref_func &&
-                 std::all_of(elem_exprs.begin(), elem_exprs.end(),
-                             [](const ElemExpr& elem_expr) {
-                               return elem_expr.kind == ElemExprKind::RefFunc;
-                             });
+  all_ref_func =
+      all_ref_func &&
+      std::all_of(elem_exprs.begin(), elem_exprs.end(),
+                  [](const ExprList& elem_expr) {
+                    return elem_expr.front().type() == ExprType::RefFunc;
+                  });
   if (!all_ref_func) {
     flags |= SegUseElemExprs;
   }
