@@ -188,6 +188,13 @@ struct FuncSignature {
   TypeVector param_types;
   TypeVector result_types;
 
+  // Some types can have names, for example (ref $foo) has type $foo.
+  // So to use this type we need to translate its name into
+  // a proper index from the module type section.
+  // This is the mapping from parameter/result index to its name.
+  std::unordered_map<uint32_t, std::string> param_type_names;
+  std::unordered_map<uint32_t, std::string> result_type_names;
+
   Index GetNumParams() const { return param_types.size(); }
   Index GetNumResults() const { return result_types.size(); }
   Type GetParamType(Index index) const { return param_types[index]; }
@@ -292,10 +299,10 @@ enum class ExprType {
   Block,
   Br,
   BrIf,
-  BrOnExn,
   BrTable,
   Call,
   CallIndirect,
+  CallRef,
   Compare,
   Const,
   Convert,
@@ -324,8 +331,11 @@ enum class ExprType {
   ReturnCallIndirect,
   Select,
   SimdLaneOp,
+  SimdLoadLane,
+  SimdStoreLane,
   SimdShuffleOp,
   LoadSplat,
+  LoadZero,
   Store,
   TableCopy,
   ElemDrop,
@@ -362,6 +372,25 @@ struct Block {
   Location end_loc;
 };
 
+struct Catch {
+  explicit Catch(const Location& loc = Location()) : loc(loc) {}
+  explicit Catch(const Var& var, const Location& loc = Location())
+      : loc(loc), var(var) {}
+  Location loc;
+  Var var;
+  ExprList exprs;
+  bool IsCatchAll() const {
+    return var.is_index() && var.index() == kInvalidIndex;
+  }
+};
+typedef std::vector<Catch> CatchVector;
+
+enum class TryKind {
+  Plain,
+  Catch,
+  Delegate
+};
+
 class Expr : public intrusive_list_base<Expr> {
  public:
   WABT_DISALLOW_COPY_AND_ASSIGN(Expr);
@@ -389,16 +418,40 @@ class ExprMixin : public Expr {
   explicit ExprMixin(const Location& loc = Location()) : Expr(TypeEnum, loc) {}
 };
 
+template <ExprType TypeEnum>
+class MemoryExpr : public ExprMixin<TypeEnum> {
+ public:
+  MemoryExpr(Var memidx, const Location& loc = Location())
+      : ExprMixin<TypeEnum>(loc), memidx(memidx) {}
+
+  Var memidx;
+};
+
+template <ExprType TypeEnum>
+class MemoryBinaryExpr : public ExprMixin<TypeEnum> {
+ public:
+  MemoryBinaryExpr(Var srcmemidx,
+                   Var destmemidx,
+                   const Location& loc = Location())
+      : ExprMixin<TypeEnum>(loc),
+        srcmemidx(srcmemidx),
+        destmemidx(destmemidx) {}
+
+  Var srcmemidx;
+  Var destmemidx;
+};
+
 typedef ExprMixin<ExprType::Drop> DropExpr;
-typedef ExprMixin<ExprType::MemoryGrow> MemoryGrowExpr;
-typedef ExprMixin<ExprType::MemorySize> MemorySizeExpr;
-typedef ExprMixin<ExprType::MemoryCopy> MemoryCopyExpr;
-typedef ExprMixin<ExprType::MemoryFill> MemoryFillExpr;
 typedef ExprMixin<ExprType::Nop> NopExpr;
-typedef ExprMixin<ExprType::Rethrow> RethrowExpr;
 typedef ExprMixin<ExprType::Return> ReturnExpr;
 typedef ExprMixin<ExprType::Unreachable> UnreachableExpr;
 
+typedef MemoryExpr<ExprType::MemoryGrow> MemoryGrowExpr;
+typedef MemoryExpr<ExprType::MemorySize> MemorySizeExpr;
+typedef MemoryExpr<ExprType::MemoryFill> MemoryFillExpr;
+
+typedef MemoryBinaryExpr<ExprType::MemoryCopy> MemoryCopyExpr;
+
 template <ExprType TypeEnum>
 class RefTypeExpr : public ExprMixin<TypeEnum> {
  public:
@@ -435,6 +488,40 @@ class SimdLaneOpExpr : public ExprMixin<ExprType::SimdLaneOp> {
   uint64_t val;
 };
 
+class SimdLoadLaneExpr : public OpcodeExpr<ExprType::SimdLoadLane> {
+ public:
+  SimdLoadLaneExpr(Opcode opcode,
+                   Address align,
+                   Address offset,
+                   uint64_t val,
+                   const Location& loc = Location())
+      : OpcodeExpr<ExprType::SimdLoadLane>(opcode, loc),
+        align(align),
+        offset(offset),
+        val(val) {}
+
+  Address align;
+  Address offset;
+  uint64_t val;
+};
+
+class SimdStoreLaneExpr : public OpcodeExpr<ExprType::SimdStoreLane> {
+ public:
+  SimdStoreLaneExpr(Opcode opcode,
+                    Address align,
+                    Address offset,
+                    uint64_t val,
+                    const Location& loc = Location())
+      : OpcodeExpr<ExprType::SimdStoreLane>(opcode, loc),
+        align(align),
+        offset(offset),
+        val(val) {}
+
+  Address align;
+  Address offset;
+  uint64_t val;
+};
+
 class SimdShuffleOpExpr : public ExprMixin<ExprType::SimdShuffleOp> {
  public:
   SimdShuffleOpExpr(Opcode opcode, v128 val, const Location& loc = Location())
@@ -453,6 +540,15 @@ class VarExpr : public ExprMixin<TypeEnum> {
   Var var;
 };
 
+template <ExprType TypeEnum>
+class MemoryVarExpr : public MemoryExpr<TypeEnum> {
+ public:
+  MemoryVarExpr(const Var& var, Var memidx, const Location& loc = Location())
+      : MemoryExpr<TypeEnum>(memidx, loc), var(var) {}
+
+  Var var;
+};
+
 typedef VarExpr<ExprType::Br> BrExpr;
 typedef VarExpr<ExprType::BrIf> BrIfExpr;
 typedef VarExpr<ExprType::Call> CallExpr;
@@ -464,8 +560,8 @@ typedef VarExpr<ExprType::LocalSet> LocalSetExpr;
 typedef VarExpr<ExprType::LocalTee> LocalTeeExpr;
 typedef VarExpr<ExprType::ReturnCall> ReturnCallExpr;
 typedef VarExpr<ExprType::Throw> ThrowExpr;
+typedef VarExpr<ExprType::Rethrow> RethrowExpr;
 
-typedef VarExpr<ExprType::MemoryInit> MemoryInitExpr;
 typedef VarExpr<ExprType::DataDrop> DataDropExpr;
 typedef VarExpr<ExprType::ElemDrop> ElemDropExpr;
 typedef VarExpr<ExprType::TableGet> TableGetExpr;
@@ -474,6 +570,8 @@ typedef VarExpr<ExprType::TableGrow> TableGrowExpr;
 typedef VarExpr<ExprType::TableSize> TableSizeExpr;
 typedef VarExpr<ExprType::TableFill> TableFillExpr;
 
+typedef MemoryVarExpr<ExprType::MemoryInit> MemoryInitExpr;
+
 class SelectExpr : public ExprMixin<ExprType::Select> {
  public:
   SelectExpr(TypeVector type, const Location& loc = Location())
@@ -523,6 +621,16 @@ class ReturnCallIndirectExpr : public ExprMixin<ExprType::ReturnCallIndirect> {
   Var table;
 };
 
+class CallRefExpr : public ExprMixin<ExprType::CallRef> {
+ public:
+  explicit CallRefExpr(const Location &loc = Location())
+      : ExprMixin<ExprType::CallRef>(loc) {}
+
+  // This field is setup only during Validate phase,
+  // so keep that in mind when you use it.
+  Var function_type_index;
+};
+
 template <ExprType TypeEnum>
 class BlockExprBase : public ExprMixin<TypeEnum> {
  public:
@@ -548,19 +656,12 @@ class IfExpr : public ExprMixin<ExprType::If> {
 class TryExpr : public ExprMixin<ExprType::Try> {
  public:
   explicit TryExpr(const Location& loc = Location())
-      : ExprMixin<ExprType::Try>(loc) {}
+      : ExprMixin<ExprType::Try>(loc), kind(TryKind::Plain) {}
 
+  TryKind kind;
   Block block;
-  ExprList catch_;
-};
-
-class BrOnExnExpr : public ExprMixin<ExprType::BrOnExn> {
- public:
-  BrOnExnExpr(const Location& loc = Location())
-      : ExprMixin<ExprType::BrOnExn>(loc) {}
-
-  Var label_var;
-  Var event_var;
+  CatchVector catches;
+  Var delegate_target;
 };
 
 class BrTableExpr : public ExprMixin<ExprType::BrTable> {
@@ -598,8 +699,27 @@ class LoadStoreExpr : public ExprMixin<TypeEnum> {
   Address offset;
 };
 
-typedef LoadStoreExpr<ExprType::Load> LoadExpr;
-typedef LoadStoreExpr<ExprType::Store> StoreExpr;
+template <ExprType TypeEnum>
+class MemoryLoadStoreExpr : public MemoryExpr<TypeEnum> {
+ public:
+  MemoryLoadStoreExpr(Opcode opcode,
+                      Var memidx,
+                      Address align,
+                      Address offset,
+                      const Location& loc = Location())
+      : MemoryExpr<TypeEnum>(memidx, loc),
+        opcode(opcode),
+        align(align),
+        offset(offset) {}
+
+  Opcode opcode;
+  Address align;
+  Address offset;
+};
+
+typedef MemoryLoadStoreExpr<ExprType::Load> LoadExpr;
+typedef MemoryLoadStoreExpr<ExprType::Store> StoreExpr;
+
 typedef LoadStoreExpr<ExprType::AtomicLoad> AtomicLoadExpr;
 typedef LoadStoreExpr<ExprType::AtomicStore> AtomicStoreExpr;
 typedef LoadStoreExpr<ExprType::AtomicRmw> AtomicRmwExpr;
@@ -607,6 +727,7 @@ typedef LoadStoreExpr<ExprType::AtomicRmwCmpxchg> AtomicRmwCmpxchgExpr;
 typedef LoadStoreExpr<ExprType::AtomicWait> AtomicWaitExpr;
 typedef LoadStoreExpr<ExprType::AtomicNotify> AtomicNotifyExpr;
 typedef LoadStoreExpr<ExprType::LoadSplat> LoadSplatExpr;
+typedef LoadStoreExpr<ExprType::LoadZero> LoadZeroExpr;
 
 class AtomicFenceExpr : public ExprMixin<ExprType::AtomicFence> {
  public:
@@ -618,8 +739,8 @@ class AtomicFenceExpr : public ExprMixin<ExprType::AtomicFence> {
   uint32_t consistency_model;
 };
 
-struct Event {
-  explicit Event(string_view name) : name(name.to_string()) {}
+struct Tag {
+  explicit Tag(string_view name) : name(name.to_string()) {}
 
   std::string name;
   FuncDeclaration decl;
@@ -726,22 +847,7 @@ struct Table {
   Type elem_type;
 };
 
-enum class ElemExprKind {
-  RefNull,
-  RefFunc,
-};
-
-struct ElemExpr {
-  ElemExpr() : kind(ElemExprKind::RefNull), type(Type::FuncRef) {}
-  explicit ElemExpr(Var var) : kind(ElemExprKind::RefFunc), var(var) {}
-  explicit ElemExpr(Type type) : kind(ElemExprKind::RefNull), type(type) {}
-
-  ElemExprKind kind;
-  Var var;    // Only used when kind == RefFunc.
-  Type type;  // Only used when kind == RefNull
-};
-
-typedef std::vector<ElemExpr> ElemExprVector;
+typedef std::vector<ExprList> ExprListVector;
 
 struct ElemSegment {
   explicit ElemSegment(string_view name) : name(name.to_string()) {}
@@ -752,7 +858,7 @@ struct ElemSegment {
   Var table_var;
   Type elem_type;
   ExprList offset;
-  ElemExprVector elem_exprs;
+  ExprListVector elem_exprs;
 };
 
 struct Memory {
@@ -832,12 +938,12 @@ class GlobalImport : public ImportMixin<ExternalKind::Global> {
   Global global;
 };
 
-class EventImport : public ImportMixin<ExternalKind::Event> {
+class TagImport : public ImportMixin<ExternalKind::Tag> {
  public:
-  explicit EventImport(string_view name = string_view())
-      : ImportMixin<ExternalKind::Event>(), event(name) {}
+  explicit TagImport(string_view name = string_view())
+      : ImportMixin<ExternalKind::Tag>(), tag(name) {}
 
-  Event event;
+  Tag tag;
 };
 
 struct Export {
@@ -857,7 +963,7 @@ enum class ModuleFieldType {
   Memory,
   DataSegment,
   Start,
-  Event
+  Tag
 };
 
 class ModuleField : public intrusive_list_base<ModuleField> {
@@ -975,13 +1081,13 @@ class DataSegmentModuleField
   DataSegment data_segment;
 };
 
-class EventModuleField : public ModuleFieldMixin<ModuleFieldType::Event> {
+class TagModuleField : public ModuleFieldMixin<ModuleFieldType::Tag> {
  public:
-  explicit EventModuleField(const Location& loc = Location(),
-                            string_view name = string_view())
-      : ModuleFieldMixin<ModuleFieldType::Event>(loc), event(name) {}
+  explicit TagModuleField(const Location& loc = Location(),
+                          string_view name = string_view())
+      : ModuleFieldMixin<ModuleFieldType::Tag>(loc), tag(name) {}
 
-  Event event;
+  Tag tag;
 };
 
 class StartModuleField : public ModuleFieldMixin<ModuleFieldType::Start> {
@@ -1011,8 +1117,8 @@ struct Module {
   const Global* GetGlobal(const Var&) const;
   Global* GetGlobal(const Var&);
   const Export* GetExport(string_view) const;
-  Event* GetEvent(const Var&) const;
-  Index GetEventIndex(const Var&) const;
+  Tag* GetTag(const Var&) const;
+  Index GetTagIndex(const Var&) const;
   const DataSegment* GetDataSegment(const Var&) const;
   DataSegment* GetDataSegment(const Var&);
   Index GetDataSegmentIndex(const Var&) const;
@@ -1028,7 +1134,7 @@ struct Module {
   // TODO(binji): move this into a builder class?
   void AppendField(std::unique_ptr<DataSegmentModuleField>);
   void AppendField(std::unique_ptr<ElemSegmentModuleField>);
-  void AppendField(std::unique_ptr<EventModuleField>);
+  void AppendField(std::unique_ptr<TagModuleField>);
   void AppendField(std::unique_ptr<ExportModuleField>);
   void AppendField(std::unique_ptr<FuncModuleField>);
   void AppendField(std::unique_ptr<TypeModuleField>);
@@ -1044,7 +1150,7 @@ struct Module {
   std::string name;
   ModuleFieldList fields;
 
-  Index num_event_imports = 0;
+  Index num_tag_imports = 0;
   Index num_func_imports = 0;
   Index num_table_imports = 0;
   Index num_memory_imports = 0;
@@ -1052,7 +1158,7 @@ struct Module {
 
   // Cached for convenience; the pointers are shared with values that are
   // stored in either ModuleField or Import.
-  std::vector<Event*> events;
+  std::vector<Tag*> tags;
   std::vector<Func*> funcs;
   std::vector<Global*> globals;
   std::vector<Import*> imports;
@@ -1064,7 +1170,7 @@ struct Module {
   std::vector<DataSegment*> data_segments;
   std::vector<Var*> starts;
 
-  BindingHash event_bindings;
+  BindingHash tag_bindings;
   BindingHash func_bindings;
   BindingHash global_bindings;
   BindingHash export_bindings;
@@ -1190,9 +1296,10 @@ enum class CommandType {
   AssertReturn,
   AssertTrap,
   AssertExhaustion,
+  AssertException,
 
   First = Module,
-  Last = AssertExhaustion,
+  Last = AssertException,
 };
 static const int kCommandTypeCount = WABT_ENUM_COUNT(CommandType);
 
@@ -1269,6 +1376,12 @@ typedef AssertModuleCommand<CommandType::AssertUnlinkable>
 typedef AssertModuleCommand<CommandType::AssertUninstantiable>
     AssertUninstantiableCommand;
 
+class AssertExceptionCommand
+    : public CommandMixin<CommandType::AssertException> {
+ public:
+  ActionPtr action;
+};
+
 typedef std::unique_ptr<Command> CommandPtr;
 typedef std::vector<CommandPtr> CommandPtrVector;
 
