@@ -30,8 +30,8 @@ class InterpTest : public ::testing::Test {
   void ReadModule(const std::vector<u8>& data) {
     Errors errors;
     ReadBinaryOptions options;
-    Result result = ReadBinaryInterp(data.data(), data.size(), options, &errors,
-                                     &module_desc_);
+    Result result = ReadBinaryInterp("<internal>", data.data(), data.size(),
+                                     options, &errors, &module_desc_);
     ASSERT_EQ(Result::Ok, result)
         << FormatErrorsToString(errors, Location::Type::Binary);
   }
@@ -102,7 +102,7 @@ TEST_F(InterpTest, MVP) {
   EXPECT_EQ(1u, module_desc_.tables.size());
   EXPECT_EQ(1u, module_desc_.memories.size());
   EXPECT_EQ(1u, module_desc_.globals.size());
-  EXPECT_EQ(0u, module_desc_.events.size());
+  EXPECT_EQ(0u, module_desc_.tags.size());
   EXPECT_EQ(1u, module_desc_.exports.size());
   EXPECT_EQ(1u, module_desc_.starts.size());
   EXPECT_EQ(1u, module_desc_.elems.size());
