@@ -70,6 +70,54 @@ Result TypeChecker::GetLabel(Index depth, Label** out_label) {
   return Result::Ok;
 }
 
+Result TypeChecker::GetRethrowLabel(Index depth, Label** out_label) {
+  if (Failed(GetLabel(depth, out_label))) {
+    return Result::Error;
+  }
+
+  if ((*out_label)->label_type == LabelType::Catch) {
+    return Result::Ok;
+  }
+
+  std::string candidates;
+  for (Index idx = 0; idx < label_stack_.size(); idx++) {
+    LabelType type = label_stack_[label_stack_.size() - idx - 1].label_type;
+    if (type == LabelType::Catch) {
+      if (!candidates.empty()) {
+        candidates.append(", ");
+      }
+      candidates.append(std::to_string(idx));
+    }
+  }
+
+  if (candidates.empty()) {
+    PrintError("rethrow not in try catch block");
+  } else {
+    PrintError("invalid rethrow depth: %" PRIindex " (catches: %s)", depth,
+               candidates.c_str());
+  }
+  *out_label = nullptr;
+  return Result::Error;
+}
+
+Result TypeChecker::GetCatchCount(Index depth, Index* out_count) {
+  Label* unused;
+  if (Failed(GetLabel(depth, &unused))) {
+    return Result::Error;
+  }
+
+  Index catch_count = 0;
+  for (Index idx = 0; idx <= depth; idx++) {
+    LabelType type = label_stack_[label_stack_.size() - idx - 1].label_type;
+    if (type == LabelType::Catch) {
+      catch_count++;
+    }
+  }
+  *out_count = catch_count;
+
+  return Result::Ok;
+}
+
 Result TypeChecker::TopLabel(Label** out_label) {
   return GetLabel(0, out_label);
 }
@@ -110,6 +158,13 @@ Result TypeChecker::CheckLabelType(Label* label, LabelType label_type) {
   return label->label_type == label_type ? Result::Ok : Result::Error;
 }
 
+Result TypeChecker::Check2LabelTypes(Label* label,
+                                     LabelType label_type1,
+                                     LabelType label_type2) {
+  return label->label_type == label_type1 ||
+         label->label_type == label_type2 ? Result::Ok : Result::Error;
+}
+
 Result TypeChecker::GetThisFunctionLabel(Label** label) {
   return GetLabel(label_stack_.size() - 1, label);
 }
@@ -161,7 +216,7 @@ Result TypeChecker::CheckTypeStackEnd(const char* desc) {
   Result result = (type_stack_.size() == label->type_stack_limit)
                       ? Result::Ok
                       : Result::Error;
-  PrintStackIfFailed(result, desc);
+  PrintStackIfFailedV(result, desc, {}, /*is_end=*/true);
   return result;
 }
 
@@ -169,6 +224,12 @@ Result TypeChecker::CheckType(Type actual, Type expected) {
   if (expected == Type::Any || actual == Type::Any) {
     return Result::Ok;
   }
+
+  if (expected == Type::Reference && actual == Type::Reference) {
+    return expected.GetReferenceIndex() == actual.GetReferenceIndex()
+               ? Result::Ok
+               : Result::Error;
+  }
   if (actual != expected) {
     return Result::Error;
   }
@@ -295,9 +356,10 @@ Result TypeChecker::CheckOpcode3(Opcode opcode,
   return result;
 }
 
-void TypeChecker::PrintStackIfFailed(Result result,
-                                     const char* desc,
-                                     const TypeVector& expected) {
+void TypeChecker::PrintStackIfFailedV(Result result,
+                                      const char* desc,
+                                      const TypeVector& expected,
+                                      bool is_end) {
   if (Succeeded(result)) {
     return;
   }
@@ -334,6 +396,9 @@ void TypeChecker::PrintStackIfFailed(Result result,
   }
 
   std::string message = "type mismatch in ";
+  if (is_end) {
+    message = "type mismatch at end of ";
+  }
   message += desc;
   message += ", expected ";
   message += TypesToString(expected);
@@ -408,20 +473,6 @@ Result TypeChecker::OnBrIf(Index depth) {
   return result;
 }
 
-Result TypeChecker::OnBrOnExn(Index depth, const TypeVector& types) {
-  Result result = PopAndCheck1Type(Type::ExnRef, "br_on_exn");
-  Label* label;
-  CHECK_RESULT(GetLabel(depth, &label));
-  if (Failed(CheckTypes(types, label->br_types()))) {
-    PrintError("br_on_exn has inconsistent types: expected %s, got %s",
-               TypesToString(label->br_types()).c_str(),
-               TypesToString(types).c_str());
-    result = Result::Error;
-  }
-  PushType(Type::ExnRef);
-  return result;
-}
-
 Result TypeChecker::BeginBrTable() {
   br_table_sig_ = nullptr;
   return PopAndCheck1Type(Type::I32, "br_table");
@@ -466,6 +517,27 @@ Result TypeChecker::OnCallIndirect(const TypeVector& param_types,
   return result;
 }
 
+Result TypeChecker::OnIndexedFuncRef(Index* out_index) {
+  Type type;
+  Result result = PeekType(0, &type);
+  if (!(type == Type::Any || type.IsReferenceWithIndex())) {
+    TypeVector actual;
+    if (Succeeded(result)) {
+      actual.push_back(type);
+    }
+    std::string message =
+        "type mismatch in call_ref, expected reference but got " +
+        TypesToString(actual);
+    PrintError("%s", message.c_str());
+    result = Result::Error;
+  }
+  if (Succeeded(result)) {
+    *out_index = type.GetReferenceIndex();
+  }
+  result |= DropTypes(1);
+  return result;
+}
+
 Result TypeChecker::OnReturnCall(const TypeVector& param_types,
                                  const TypeVector& result_types) {
   Result result = PopAndCheckSignature(param_types, "return_call");
@@ -496,17 +568,17 @@ Result TypeChecker::OnCompare(Opcode opcode) {
   return CheckOpcode2(opcode);
 }
 
-Result TypeChecker::OnCatch() {
+Result TypeChecker::OnCatch(const TypeVector& sig) {
   Result result = Result::Ok;
   Label* label;
   CHECK_RESULT(TopLabel(&label));
-  result |= CheckLabelType(label, LabelType::Try);
+  result |= Check2LabelTypes(label, LabelType::Try, LabelType::Catch);
   result |= PopAndCheckSignature(label->result_types, "try block");
   result |= CheckTypeStackEnd("try block");
   ResetTypeStackToLabel(label);
   label->label_type = LabelType::Catch;
   label->unreachable = false;
-  PushType(Type::ExnRef);
+  PushTypes(sig);
   return result;
 }
 
@@ -519,6 +591,27 @@ Result TypeChecker::OnConvert(Opcode opcode) {
   return CheckOpcode1(opcode);
 }
 
+Result TypeChecker::OnDelegate(Index depth) {
+  Result result = Result::Ok;
+  Label* label;
+  // Delegate starts counting after the current try, as the delegate
+  // instruction is not actually in the try block.
+  CHECK_RESULT(GetLabel(depth + 1, &label));
+
+  Label* try_label;
+  CHECK_RESULT(TopLabel(&try_label));
+  result |= CheckLabelType(try_label, LabelType::Try);
+  result |= PopAndCheckSignature(try_label->result_types, "try block");
+  result |= CheckTypeStackEnd("try block");
+  ResetTypeStackToLabel(try_label);
+
+  // Since an end instruction does not follow a delegate, we push
+  // the block results here and pop the label.
+  PushTypes(try_label->result_types);
+  PopLabel();
+  return result;
+}
+
 Result TypeChecker::OnDrop() {
   Result result = Result::Ok;
   result |= DropTypes(1);
@@ -531,8 +624,8 @@ Result TypeChecker::OnElse() {
   Label* label;
   CHECK_RESULT(TopLabel(&label));
   result |= CheckLabelType(label, LabelType::If);
-  result |= PopAndCheckSignature(label->result_types, "if true branch");
-  result |= CheckTypeStackEnd("if true branch");
+  result |= PopAndCheckSignature(label->result_types, "`if true` branch");
+  result |= CheckTypeStackEnd("`if true` branch");
   ResetTypeStackToLabel(label);
   PushTypes(label->param_types);
   label->label_type = LabelType::Else;
@@ -555,7 +648,8 @@ Result TypeChecker::OnEnd(Label* label,
 Result TypeChecker::OnEnd() {
   Result result = Result::Ok;
   static const char* s_label_type_name[] = {
-      "function", "block", "loop", "if", "if false branch", "try", "try catch"};
+      "function", "initializer expression", "block", "loop",
+      "if",       "`if false` branch",      "try",   "try catch"};
   WABT_STATIC_ASSERT(WABT_ARRAY_SIZE(s_label_type_name) == kLabelTypeCount);
   Label* label;
   CHECK_RESULT(TopLabel(&label));
@@ -566,6 +660,7 @@ Result TypeChecker::OnEnd() {
     // an empty else block.
     CHECK_RESULT(OnElse());
   }
+
   const char* desc = s_label_type_name[static_cast<int>(label->label_type)];
   result |= OnEnd(label, desc, desc);
   return result;
@@ -638,7 +733,7 @@ Result TypeChecker::OnMemoryInit(uint32_t segment, const Limits& limits) {
 }
 
 Result TypeChecker::OnMemorySize(const Limits& limits) {
-  PushType(limits.is_64 ? Type::I64 : Type::I32);
+  PushType(limits.IndexType());
   return Result::Ok;
 }
 
@@ -679,8 +774,12 @@ Result TypeChecker::OnTableFill(Type elem_type) {
   return PopAndCheck3Types(Type::I32, elem_type, Type::I32, "table.fill");
 }
 
-Result TypeChecker::OnRefFuncExpr(Index) {
-  PushType(Type::FuncRef);
+Result TypeChecker::OnRefFuncExpr(Index func_type) {
+  if (features_.function_references_enabled()) {
+    PushType(Type(Type::Reference, func_type));
+  } else {
+    PushType(Type::FuncRef);
+  }
   return Result::Ok;
 }
 
@@ -692,7 +791,7 @@ Result TypeChecker::OnRefNullExpr(Type type) {
 Result TypeChecker::OnRefIsNullExpr() {
   Type type;
   Result result = PeekType(0, &type);
-  if (!type.IsRef()) {
+  if (!(type == Type::Any || type.IsRef())) {
     TypeVector actual;
     if (Succeeded(result)) {
       actual.push_back(type);
@@ -708,8 +807,10 @@ Result TypeChecker::OnRefIsNullExpr() {
   return result;
 }
 
-Result TypeChecker::OnRethrow() {
-  Result result = PopAndCheck1Type(Type::ExnRef, "rethrow");
+Result TypeChecker::OnRethrow(Index depth) {
+  Result result = Result::Ok;
+  Label* label;
+  CHECK_RESULT(GetRethrowLabel(depth, &label));
   CHECK_RESULT(SetUnreachable());
   return result;
 }
@@ -730,7 +831,7 @@ Result TypeChecker::OnReturn() {
   return result;
 }
 
-Result TypeChecker::OnSelect(Type expected) {
+Result TypeChecker::OnSelect(const TypeVector& expected) {
   Result result = Result::Ok;
   Type type1 = Type::Any;
   Type type2 = Type::Any;
@@ -738,7 +839,7 @@ Result TypeChecker::OnSelect(Type expected) {
   result |= PeekAndCheckType(0, Type::I32);
   result |= PeekType(1, &type1);
   result |= PeekType(2, &type2);
-  if (expected == Type::Void) {
+  if (expected.empty()) {
     if (type1.IsRef() || type2.IsRef()) {
       result = Result::Error;
     } else {
@@ -746,8 +847,9 @@ Result TypeChecker::OnSelect(Type expected) {
       result_type = type1;
     }
   } else {
-    result |= CheckType(type1, expected);
-    result |= CheckType(type2, expected);
+    assert(expected.size() == 1);
+    result |= CheckType(type1, expected[0]);
+    result |= CheckType(type2, expected[0]);
   }
   PrintStackIfFailed(result, "select", result_type, result_type, Type::I32);
   result |= DropTypes(3);
@@ -809,6 +911,30 @@ Result TypeChecker::OnSimdLaneOp(Opcode opcode, uint64_t lane_idx) {
   return result;
 }
 
+Result TypeChecker::OnSimdLoadLane(Opcode opcode, const Limits& limits, uint64_t lane_idx) {
+  Result result = Result::Ok;
+  uint32_t lane_count = opcode.GetSimdLaneCount();
+  if (lane_idx >= lane_count) {
+    PrintError("lane index must be less than %d (got %" PRIu64 ")", lane_count,
+               lane_idx);
+    result = Result::Error;
+  }
+  result |= CheckOpcode2(opcode, &limits);
+  return result;
+}
+
+Result TypeChecker::OnSimdStoreLane(Opcode opcode, const Limits& limits, uint64_t lane_idx) {
+  Result result = Result::Ok;
+  uint32_t lane_count = opcode.GetSimdLaneCount();
+  if (lane_idx >= lane_count) {
+    PrintError("lane index must be less than %d (got %" PRIu64 ")", lane_count,
+               lane_idx);
+    result = Result::Error;
+  }
+  result |= CheckOpcode2(opcode, &limits);
+  return result;
+}
+
 Result TypeChecker::OnSimdShuffleOp(Opcode opcode, v128 lane_idx) {
   Result result = Result::Ok;
   uint8_t simd_data[16];
@@ -837,4 +963,20 @@ Result TypeChecker::EndFunction() {
   return result;
 }
 
+Result TypeChecker::BeginInitExpr(Type type) {
+  type_stack_.clear();
+  label_stack_.clear();
+  PushLabel(LabelType::InitExpr, TypeVector(), {type});
+  return Result::Ok;
+}
+
+Result TypeChecker::EndInitExpr() {
+  Result result = Result::Ok;
+  Label* label;
+  CHECK_RESULT(TopLabel(&label));
+  result |= CheckLabelType(label, LabelType::InitExpr);
+  result |= OnEnd(label, "initializer expression", "initializer expression");
+  return result;
+}
+
 }  // namespace wabt
