@@ -90,10 +90,10 @@ class Validator : public ExprVisitor::Delegate {
   Result EndBlockExpr(BlockExpr*) override;
   Result OnBrExpr(BrExpr*) override;
   Result OnBrIfExpr(BrIfExpr*) override;
-  Result OnBrOnExnExpr(BrOnExnExpr*) override;
   Result OnBrTableExpr(BrTableExpr*) override;
   Result OnCallExpr(CallExpr*) override;
   Result OnCallIndirectExpr(CallIndirectExpr*) override;
+  Result OnCallRefExpr(CallRefExpr*) override;
   Result OnCompareExpr(CompareExpr*) override;
   Result OnConstExpr(ConstExpr*) override;
   Result OnConvertExpr(ConvertExpr*) override;
@@ -135,7 +135,8 @@ class Validator : public ExprVisitor::Delegate {
   Result OnUnaryExpr(UnaryExpr*) override;
   Result OnUnreachableExpr(UnreachableExpr*) override;
   Result BeginTryExpr(TryExpr*) override;
-  Result OnCatchExpr(TryExpr*) override;
+  Result OnCatchExpr(TryExpr*, Catch*) override;
+  Result OnDelegateExpr(TryExpr*) override;
   Result EndTryExpr(TryExpr*) override;
   Result OnThrowExpr(ThrowExpr*) override;
   Result OnRethrowExpr(RethrowExpr*) override;
@@ -148,8 +149,11 @@ class Validator : public ExprVisitor::Delegate {
   Result OnAtomicRmwCmpxchgExpr(AtomicRmwCmpxchgExpr*) override;
   Result OnTernaryExpr(TernaryExpr*) override;
   Result OnSimdLaneOpExpr(SimdLaneOpExpr*) override;
+  Result OnSimdLoadLaneExpr(SimdLoadLaneExpr*) override;
+  Result OnSimdStoreLaneExpr(SimdStoreLaneExpr*) override;
   Result OnSimdShuffleOpExpr(SimdShuffleOpExpr*) override;
   Result OnLoadSplatExpr(LoadSplatExpr*) override;
+  Result OnLoadZeroExpr(LoadZeroExpr*) override;
 
  private:
   Type GetDeclarationType(const FuncDeclaration&);
@@ -182,7 +186,8 @@ void ScriptValidator::CheckTypeIndex(const Location* loc,
   if (Failed(TypeChecker::CheckType(actual, expected))) {
     PrintError(loc,
                "type mismatch for %s %" PRIindex " of %s. got %s, expected %s",
-               index_kind, index, desc, actual.GetName(), expected.GetName());
+               index_kind, index, desc, actual.GetName().c_str(),
+               expected.GetName().c_str());
   }
 }
 
@@ -249,12 +254,6 @@ Result Validator::OnBrIfExpr(BrIfExpr* expr) {
   return Result::Ok;
 }
 
-Result Validator::OnBrOnExnExpr(BrOnExnExpr* expr) {
-  result_ |= validator_.OnBrOnExn(expr->loc, expr->label_var,
-                                  expr->event_var);
-  return Result::Ok;
-}
-
 Result Validator::OnBrTableExpr(BrTableExpr* expr) {
   result_ |= validator_.BeginBrTable(expr->loc);
   for (const Var& var : expr->targets) {
@@ -276,6 +275,17 @@ Result Validator::OnCallIndirectExpr(CallIndirectExpr* expr) {
   return Result::Ok;
 }
 
+Result Validator::OnCallRefExpr(CallRefExpr* expr) {
+  Index function_type_index;
+  result_ |= validator_.OnCallRef(expr->loc, &function_type_index);
+  if (Succeeded(result_)) {
+    expr->function_type_index = Var{function_type_index};
+    return Result::Ok;
+  }
+
+  return Result::Error;
+}
+
 Result Validator::OnCompareExpr(CompareExpr* expr) {
   result_ |= validator_.OnCompare(expr->loc, expr->opcode);
   return Result::Ok;
@@ -325,7 +335,7 @@ Result Validator::EndIfExpr(IfExpr* expr) {
 }
 
 Result Validator::OnLoadExpr(LoadExpr* expr) {
-  result_ |= validator_.OnLoad(expr->loc, expr->opcode,
+  result_ |= validator_.OnLoad(expr->loc, expr->opcode, expr->memidx,
                                expr->opcode.GetAlignment(expr->align));
   return Result::Ok;
 }
@@ -356,7 +366,8 @@ Result Validator::EndLoopExpr(LoopExpr* expr) {
 }
 
 Result Validator::OnMemoryCopyExpr(MemoryCopyExpr* expr) {
-  result_ |= validator_.OnMemoryCopy(expr->loc);
+  result_ |=
+      validator_.OnMemoryCopy(expr->loc, expr->srcmemidx, expr->destmemidx);
   return Result::Ok;
 }
 
@@ -366,22 +377,22 @@ Result Validator::OnDataDropExpr(DataDropExpr* expr) {
 }
 
 Result Validator::OnMemoryFillExpr(MemoryFillExpr* expr) {
-  result_ |= validator_.OnMemoryFill(expr->loc);
+  result_ |= validator_.OnMemoryFill(expr->loc, expr->memidx);
   return Result::Ok;
 }
 
 Result Validator::OnMemoryGrowExpr(MemoryGrowExpr* expr) {
-  result_ |= validator_.OnMemoryGrow(expr->loc);
+  result_ |= validator_.OnMemoryGrow(expr->loc, expr->memidx);
   return Result::Ok;
 }
 
 Result Validator::OnMemoryInitExpr(MemoryInitExpr* expr) {
-  result_ |= validator_.OnMemoryInit(expr->loc, expr->var);
+  result_ |= validator_.OnMemoryInit(expr->loc, expr->var, expr->memidx);
   return Result::Ok;
 }
 
 Result Validator::OnMemorySizeExpr(MemorySizeExpr* expr) {
-  result_ |= validator_.OnMemorySize(expr->loc);
+  result_ |= validator_.OnMemorySize(expr->loc, expr->memidx);
   return Result::Ok;
 }
 
@@ -464,13 +475,8 @@ Result Validator::OnReturnCallIndirectExpr(ReturnCallIndirectExpr* expr) {
 }
 
 Result Validator::OnSelectExpr(SelectExpr* expr) {
-  Type result_type;
-  if (expr->result_type.empty()) {
-    result_type = Type::Void;
-  } else {
-    result_type = expr->result_type[0];
-  }
-  result_ |= validator_.OnSelect(expr->loc, result_type);
+  result_ |= validator_.OnSelect(expr->loc, expr->result_type.size(),
+                                 expr->result_type.data());
   // TODO: Existing behavior fails when select fails.
 #if 0
   return Result::Ok;
@@ -480,7 +486,7 @@ Result Validator::OnSelectExpr(SelectExpr* expr) {
 }
 
 Result Validator::OnStoreExpr(StoreExpr* expr) {
-  result_ |= validator_.OnStore(expr->loc, expr->opcode,
+  result_ |= validator_.OnStore(expr->loc, expr->opcode, expr->memidx,
                                 expr->opcode.GetAlignment(expr->align));
   return Result::Ok;
 }
@@ -500,8 +506,14 @@ Result Validator::BeginTryExpr(TryExpr* expr) {
   return Result::Ok;
 }
 
-Result Validator::OnCatchExpr(TryExpr* expr) {
-  result_ |= validator_.OnCatch(expr->loc);
+Result Validator::OnCatchExpr(TryExpr*, Catch* catch_) {
+  result_ |= validator_.OnCatch(catch_->loc, catch_->var,
+                                catch_->IsCatchAll());
+  return Result::Ok;
+}
+
+Result Validator::OnDelegateExpr(TryExpr* expr) {
+  result_ |= validator_.OnDelegate(expr->loc, expr->delegate_target);
   return Result::Ok;
 }
 
@@ -516,7 +528,7 @@ Result Validator::OnThrowExpr(ThrowExpr* expr) {
 }
 
 Result Validator::OnRethrowExpr(RethrowExpr* expr) {
-  result_ |= validator_.OnRethrow(expr->loc);
+  result_ |= validator_.OnRethrow(expr->loc, expr->var);
   return Result::Ok;
 }
 
@@ -571,6 +583,20 @@ Result Validator::OnSimdLaneOpExpr(SimdLaneOpExpr* expr) {
   return Result::Ok;
 }
 
+Result Validator::OnSimdLoadLaneExpr(SimdLoadLaneExpr* expr) {
+  result_ |= validator_.OnSimdLoadLane(
+      expr->loc, expr->opcode, expr->opcode.GetAlignment(expr->align),
+      expr->val);
+  return Result::Ok;
+}
+
+Result Validator::OnSimdStoreLaneExpr(SimdStoreLaneExpr* expr) {
+  result_ |= validator_.OnSimdStoreLane(expr->loc, expr->opcode,
+                                        expr->opcode.GetAlignment(expr->align),
+                                        expr->val);
+  return Result::Ok;
+}
+
 Result Validator::OnSimdShuffleOpExpr(SimdShuffleOpExpr* expr) {
   result_ |= validator_.OnSimdShuffleOp(expr->loc, expr->opcode, expr->val);
   return Result::Ok;
@@ -582,6 +608,12 @@ Result Validator::OnLoadSplatExpr(LoadSplatExpr* expr) {
   return Result::Ok;
 }
 
+Result Validator::OnLoadZeroExpr(LoadZeroExpr* expr) {
+  result_ |= validator_.OnLoadZero(expr->loc, expr->opcode,
+                                   expr->opcode.GetAlignment(expr->align));
+  return Result::Ok;
+}
+
 Validator::Validator(Errors* errors,
                      const Module* module,
                      const ValidateOptions& options)
@@ -599,11 +631,12 @@ Result Validator::CheckModule() {
       switch (f->type->kind()) {
         case TypeEntryKind::Func: {
           FuncType* func_type = cast<FuncType>(f->type.get());
-          result_ |= validator_.OnFuncType(field.loc,
-                                           func_type->sig.param_types.size(),
-                                           func_type->sig.param_types.data(),
-                                           func_type->sig.result_types.size(),
-                                           func_type->sig.result_types.data());
+          result_ |= validator_.OnFuncType(
+              field.loc, func_type->sig.param_types.size(),
+              func_type->sig.param_types.data(),
+              func_type->sig.result_types.size(),
+              func_type->sig.result_types.data(),
+              module->GetFuncTypeIndex(func_type->sig));
           break;
         }
 
@@ -660,10 +693,10 @@ Result Validator::CheckModule() {
           break;
         }
 
-        case ExternalKind::Event: {
-          auto&& event = cast<EventImport>(f->import.get())->event;
-          result_ |= validator_.OnEvent(
-              field.loc, GetFuncTypeIndex(field.loc, event.decl));
+        case ExternalKind::Tag: {
+          auto&& tag = cast<TagImport>(f->import.get())->tag;
+          result_ |= validator_.OnTag(field.loc,
+                                      GetFuncTypeIndex(field.loc, tag.decl));
           break;
         }
       }
@@ -699,46 +732,20 @@ Result Validator::CheckModule() {
       result_ |=
           validator_.OnGlobal(field.loc, f->global.type, f->global.mutable_);
 
-      if (f->global.init_expr.size() == 1) {
-        const Expr* expr = &f->global.init_expr.front();
-
-        switch (expr->type()) {
-          case ExprType::Const:
-            result_ |= validator_.OnGlobalInitExpr_Const(
-                expr->loc, cast<ConstExpr>(expr)->const_.type());
-            break;
-
-          case ExprType::GlobalGet: {
-            Var var = cast<GlobalGetExpr>(expr)->var;
-            result_ |= validator_.OnGlobalInitExpr_GlobalGet(expr->loc, var);
-            break;
-          }
-
-          case ExprType::RefFunc:
-            result_ |= validator_.OnGlobalInitExpr_RefFunc(
-                expr->loc, cast<RefFuncExpr>(expr)->var);
-            break;
-
-          case ExprType::RefNull:
-            result_ |= validator_.OnGlobalInitExpr_RefNull(
-                expr->loc, cast<RefNullExpr>(expr)->type);
-            break;
-
-          default:
-            result_ |= validator_.OnGlobalInitExpr_Other(field.loc);
-            break;
-        }
-      } else {
-        result_ |= validator_.OnGlobalInitExpr_Other(field.loc);
-      }
+      // Init expr.
+      result_ |= validator_.BeginInitExpr(field.loc, f->global.type);
+      ExprVisitor visitor(this);
+      result_ |=
+          visitor.VisitExprList(const_cast<ExprList&>(f->global.init_expr));
+      result_ |= validator_.EndInitExpr();
     }
   }
 
-  // Event section.
+  // Tag section.
   for (const ModuleField& field : module->fields) {
-    if (auto* f = dyn_cast<EventModuleField>(&field)) {
-      result_ |= validator_.OnEvent(field.loc,
-                                    GetFuncTypeIndex(field.loc, f->event.decl));
+    if (auto* f = dyn_cast<TagModuleField>(&field)) {
+      result_ |=
+          validator_.OnTag(field.loc, GetFuncTypeIndex(field.loc, f->tag.decl));
     }
   }
 
@@ -766,43 +773,33 @@ Result Validator::CheckModule() {
       validator_.OnElemSegmentElemType(f->elem_segment.elem_type);
 
       // Init expr.
-      if (f->elem_segment.offset.size() == 1) {
-        const Expr* expr = &f->elem_segment.offset.front();
-
-        switch (expr->type()) {
-          case ExprType::Const:
-            result_ |= validator_.OnElemSegmentInitExpr_Const(
-                expr->loc, cast<ConstExpr>(expr)->const_.type());
-            break;
-
-          case ExprType::GlobalGet: {
-            Var var = cast<GlobalGetExpr>(expr)->var;
-            result_ |=
-                validator_.OnElemSegmentInitExpr_GlobalGet(expr->loc, var);
-            break;
-          }
-
-          default:
-            result_ |= validator_.OnElemSegmentInitExpr_Other(field.loc);
-            break;
-        }
-      } else if (f->elem_segment.offset.size() > 1) {
-        result_ |= validator_.OnElemSegmentInitExpr_Other(field.loc);
+      if (f->elem_segment.offset.size()) {
+        result_ |= validator_.BeginInitExpr(field.loc, Type::I32);
+        ExprVisitor visitor(this);
+        result_ |= visitor.VisitExprList(
+            const_cast<ExprList&>(f->elem_segment.offset));
+        result_ |= validator_.EndInitExpr();
       }
 
       // Element expr.
       for (auto&& elem_expr : f->elem_segment.elem_exprs) {
-        switch (elem_expr.kind) {
-          case ElemExprKind::RefNull:
-            // TODO: better location?
-            result_ |= validator_.OnElemSegmentElemExpr_RefNull(field.loc,
-                                                                elem_expr.type);
-            break;
-
-          case ElemExprKind::RefFunc:
-            result_ |= validator_.OnElemSegmentElemExpr_RefFunc(
-                elem_expr.var.loc, elem_expr.var);
-            break;
+        if (elem_expr.size() == 1) {
+          const Expr* expr = &elem_expr.front();
+          switch (expr->type()) {
+            case ExprType::RefNull:
+              result_ |= validator_.OnElemSegmentElemExpr_RefNull(
+                  expr->loc, cast<RefNullExpr>(expr)->type);
+              break;
+            case ExprType::RefFunc:
+              result_ |= validator_.OnElemSegmentElemExpr_RefFunc(
+                  expr->loc, cast<RefFuncExpr>(expr)->var);
+              break;
+            default:
+              result_ |= validator_.OnElemSegmentElemExpr_Other(expr->loc);
+              break;
+          }
+        } else if (elem_expr.size() > 1) {
+          result_ |= validator_.OnElemSegmentElemExpr_Other(field.loc);
         }
       }
     }
@@ -835,28 +832,18 @@ Result Validator::CheckModule() {
           field.loc, f->data_segment.memory_var, f->data_segment.kind);
 
       // Init expr.
-      if (f->data_segment.offset.size() == 1) {
-        const Expr* expr = &f->data_segment.offset.front();
-
-        switch (expr->type()) {
-          case ExprType::Const:
-            result_ |= validator_.OnDataSegmentInitExpr_Const(
-                expr->loc, cast<ConstExpr>(expr)->const_.type());
-            break;
-
-          case ExprType::GlobalGet: {
-            Var var = cast<GlobalGetExpr>(expr)->var;
-            result_ |=
-                validator_.OnDataSegmentInitExpr_GlobalGet(expr->loc, var);
-            break;
-          }
-
-          default:
-            result_ |= validator_.OnDataSegmentInitExpr_Other(field.loc);
-            break;
+      if (f->data_segment.offset.size()) {
+        Type offset_type = Type::I32;
+        Index memory_index = module->GetMemoryIndex(f->data_segment.memory_var);
+        if (memory_index < module->memories.size() &&
+            module->memories[memory_index]->page_limits.is_64) {
+          offset_type = Type::I64;
         }
-      } else if (f->data_segment.offset.size() > 1) {
-        result_ |= validator_.OnDataSegmentInitExpr_Other(field.loc);
+        result_ |= validator_.BeginInitExpr(field.loc, offset_type);
+        ExprVisitor visitor(this);
+        result_ |= visitor.VisitExprList(
+            const_cast<ExprList&>(f->data_segment.offset));
+        result_ |= validator_.EndInitExpr();
       }
     }
   }
@@ -1012,6 +999,10 @@ void ScriptValidator::CheckCommand(const Command* command) {
       // ignore result type.
       CheckAction(cast<AssertExhaustionCommand>(command)->action.get());
       break;
+    case CommandType::AssertException:
+      // ignore result type.
+      CheckAction(cast<AssertExceptionCommand>(command)->action.get());
+      break;
   }
 }
 
